/*
** Module   : KeyInput.java
** Abstract : Extended version of keyboard input event.
**
** Copyright (c) 2005-2015, Golden Code Development Corporation.
**
** -#- -I- --Date-- --JPRM-- ---------------------------Description----------------------------
** 001 NVS 20060301   @24810 Created initial version.
** 002 NVS 20060613   @27170 Instances constructed from high level events now have the key
**                           component set to -1.
** 003 NVS 20060613   @27175 Previous fix regressed readKey() method in ThinClient. Fixed.
** 004 NVS 20080502   @38200 Added eventMode variable and isSpecial() getter.
** 005 SIY 20080901   @39640 Added flag for 'real keys' (i.e. keys typed by user).
** 006 CA  20080924   @39933 Fixed H005 regression eventMode must be passed to the c'tor.
** 007 SVL 20090224   @41356 During construction keep the original key code for real keys.
** 008 SIY 20100619          Switching to own implementation of character UI.
** 009 SIY 20100804          Changes in imports, some refactorings.
** 010 SIY 20101122          Moved from CHUI package, merged necessary code from KeyEvent.
** 011 SIY 20110112          Refactoring and simplification.
** 012 GES 20150226          Added flag to note that this event caused AUTO-RETURN.
** 013 EVL 20150401          Adding new constructor to be able for MouseEvt to extend this class.
** 014 VIG 20150423          Added MENU-DROP fix to prevent key reset.
*/

package com.goldencode.p2j.ui.client.event;


import com.goldencode.p2j.ui.*;
import com.goldencode.p2j.ui.client.event.listener.*;
import com.goldencode.p2j.ui.client.widget.*;

/**
* An event which indicates that a keystroke occurred in an object. This keystroke is represented
* in two ways: as the usual KeyEvent and as a Progress-like logical action code.
*/
public class KeyInput
extends Event
{
   /** Progress like action code or KeyInput.CHAR_UNDEFINED */
   private int actionCode = CHAR_UNDEFINED;

   /** A flag which may have varying meaning depending on the event code.
    *  For now, ENTRY event uses this flag to denote an unrelated event, so
    *  that there will be no matching LEAVE event generated.
    */
   private boolean eventMode = false;
   
   /** Mark real keys typed by user. */
   private boolean isKey = false;

   /** If <code>true</code>, the keyboard event was processed and consumed. */
   private boolean consumed = false;

   /** Key associated with the event. */
   private int key;
   
   /** <code>true</code> if this event generated a downstream AUTO-RETURN. */
   private boolean causedAutoReturn = false;

   /** Undefined character constant. */
   public static final char CHAR_UNDEFINED = (char) -1;

   /**
    * Initializes the action code to the Mapped Progress key action. If it is
    * a valid action code, the original keystroke is translated into a
    * standard, compatible code which CHARVA understands.
    * @param   source
    *          widget associated with the event
    * @param   id
    *          event category
    * @param   key
    *          event code
    */
   public KeyInput(Widget source, int id, int key) 
   {
      this(source, id, key, false);
   }

   /**
    * Initializes the action code to the Mapped Progress key action. If it is
    * a valid action code, the original keystroke is translated into a
    * standard, compatible code which CHARVA understands.
    * @param   source
    *          widget associated with the event
    * @param   id
    *          event category
    * @param   key
    *          event code
    * @param   eventMode
    *          a boolean flag specific to the event
    */
   public KeyInput(Widget source, int id, int key, boolean eventMode)
   {
      this(source, id, key, eventMode, false);
   }
   
   /**
    * Initializes the action code to the Mapped Progress key action. If it is
    * a valid action code, the original keystroke is translated into a
    * standard, compatible code which CHARVA understands.
    * @param   src
    *          widget associated with the event
    * @param   id
    *          event category
    * @param   key
    *          event code
    * @param   mode
    *          a boolean flag specific to the event
    * @param   isKey
    *          mark real key typed by user
    */
   public KeyInput(Widget src, int id, int key, boolean mode, boolean isKey) 
   {
      super(src, id);
      this.key = key;
      
      this.eventMode = mode;
      this.isKey     = isKey;

      actionCode = Keyboard.keyAction(key);

      if (actionCode == CHAR_UNDEFINED)
      {
         actionCode = key;
      }
      else if (actionCode < CHAR_UNDEFINED && actionCode != Keyboard.SE_MENU_DROP)
      {
         // a valid key action is defined; keep the original key code if we
         // have a real key, or try to "guess" a hypothetical key code that
         // would correspond the high level event generated by APPLY
         // statement.
         int standardCode = isKey ? key :
                                    Keyboard.keyCodeForAction(actionCode);
         if (standardCode != CHAR_UNDEFINED)
            setKeyCode(standardCode);
         
         // do not keep key code as action code
         if (key == actionCode && actionCode < 0)
            setKeyCode(CHAR_UNDEFINED);
      }
   }

   /**
    * Constructor for MouseEvt compatibility.
    * 
    * @param   source
    *          Event source.
    * @param   id
    *          Event type ID.
    */
   public KeyInput(Widget source, int id)
   {
      super(source, id);
   }

   /**
    * Returns the integer action code.
    *
    * @return action code
    */
   public int actionCode()
   {
      return actionCode;
   }

   /**
    * Sets the action code.
    *
    * @param code
    *        new action code to be set
    */
   public void setActionCode(int code)
   {
      actionCode = code;
   }

   /**
    * Returns the special mode flag.
    *
    * @return event mode
    */
   public boolean isSpecial()
   {
      return eventMode;
   }

   /**
    * Returns the String representation of this object.
    *
    * @return text representation of this object
    */
   @Override
   public String toString() 
   {
      return ("KeyInput: key=" + keyCode() + " action=" + actionCode +
              " mode=" + (eventMode ? "special" : "regular") + 
              " source=[" + source() + "]");
   }

   /**
    * Return <code>true</code> if event represents key typed by user (as 
    * opposed to some kind of artificially generated event).
    * 
    * @return  <code>true</code> if event was generated by real user input.
    */
   public boolean isRealKey()
   {
      return isKey;
   }

   /**
    * Invoke appropriate listeners' method depending on the event type.
    * 
    * @param   listener
    *          Event listener to which event will be passed.
    */
   public void dispatch(KeyListener listener)
   {
      if (id() == EventType.KEY_PRESSED)
         listener.onKeyPressed(this);
      else if (id() == EventType.KEY_TYPED)
         listener.onKeyTyped(this);
   }

   /**
    * Mark event as consumed.
    */
   public void consume()
   {
      consumed = true;
   }

   /**
    * Get character for this event. If event is generated for non-character 
    * key event, then {@link KeyInput#CHAR_UNDEFINED} is returned.
    * 
    * @return  character.
    */
   public char keyChar()
   {
      return (id() == EventType.KEY_PRESSED) ? CHAR_UNDEFINED : (char) key;
   }

   /**
    * Get key code for this event.
    * 
    * @return  key code.
    */
   public int keyCode()
   {
      return key;
   }

   /**
    * Check if event is already consumed.
    * 
    * @return  <code>true</code> if event is consumed.
    */
   public boolean isConsumed()
   {
      return consumed;
   }

   /**
    * Set key code for this event.
    * 
    * @param   key
    *          Key code to set.
    */
   public void setKeyCode(int key)
   {
      this.key = key;
   }
   
   /**
    * Sets the flag reporting if this event generated an AUTO-RETURN.
    *
    * @param    causedAutoReturn
    *           <code>true</code> if this event generated a downstream AUTO-RETURN.
    */
   public void setCausedAutoReturn(boolean causedAutoReturn)
   {
      this.causedAutoReturn = causedAutoReturn;
   }
   
   /**
    * Reports if this event generated an AUTO-RETURN.
    *
    * @return   <code>true</code> if this event generated a downstream AUTO-RETURN.
    */
   public boolean isCausedAutoReturn()
   {
      return causedAutoReturn;
   }
}
