/*
** Module   : AbstractContainer.java
** Abstract : Base class for widget containers.
**
** Copyright (c) 2010-2015, Golden Code Development Corporation.
** ALL RIGHTS RESERVED. Use is subject to license terms.
**
**           Golden Code Development Corporation
**                      CONFIDENTIAL
**
** -#- -I- --Date-- ---------------------------------Description----------------------------------
** 001 SIY 20100629 Created initial version
** 002 SIY 20100804 Changes in imports, some refactorings.
** 003 SIY 20100831 Changes related to refactoring of FrameHelper.
** 004 SIY 20101122 Switching from KeyEvent to KeyInput.
** 005 SIY 20110425 Fixed next/prev focus handling for deeply nested 
**                  containers.
** 006 GES 20111012 Moved static data to context-local. This allows multiple
**                  client sessions to be hosted in the server.
** 007 EVL 20140528 Fix for displaying the labels for hidden widgets in a frame.  When the hidden
**                  widget become visible the label text should be drawn even later the widget
**                  will be hidden again.
** 008 VIG 20140507 Added tab order management methods and changes.
** 009 CA  20140805 Added generics to force the abstract widget classes to not use implementation-
**                  specific code: the abstract widgets will have access only to the base APIs
**                  provided by the OutputManager.
**                  Removed size API, which collided with the dimension API.
**                  First attempt at clipping support (requires decimal char units to work).
**                  First pass at mouse support.  Added widget destroy notifications.
** 010 HC  20140911 GUI coordinate support - data types refactoring.
** 011 CA  20140926 Added shared widget configuration support, to allow GUI/ChUI concrete
**                  implementation for the same widget ID.  Refactored the widget configuration
**                  classes: all fields were made public; these classes need to be as dumb as
**                  possible, all logic related to setting/getting a certain field should be at
**                  the widget or at the caller. Refs #2254
** 012 MAG 20140929 Fixed findMouseSource alghoritm.
** 013 CA  20141021 Changed findMouseSource to allow Z-order.  Fixed draw() to allow drawing of
**                  nested containers.
** 014 MAG 20141020 Added virtual size computation. Eliminated GUI drawing area settings in draw()
**                  Added register widgets.
** 015 HC  20141021 Implemented window move/position support.
** 016 CA  20141119 Refactored drawing code to enable multiple OS windows, for GUI support.  Misc
**                  changes/fixes related to drawing on multiple windows.
** 017 VIG 20150216 Added method getWidgetsList.
** 018 HC  20150221 Improvements to frame tab order.
** 019 VIG 20150302 SubMenu can gain a focus. Added exclusion.
*/

package com.goldencode.p2j.ui.client.widget;

import java.util.*;

import com.goldencode.p2j.security.ContextLocal;
import com.goldencode.p2j.ui.*;
import com.goldencode.p2j.ui.client.*;
import com.goldencode.p2j.ui.client.event.KeyInput;
import com.goldencode.p2j.ui.client.layout.LayoutManager;
import com.goldencode.p2j.util.Iterables;

/**
 * Abstract widget container.
 */
public abstract class AbstractContainer<O extends OutputManager<?>>
extends AbstractWidget<O>
implements Container<O>
{
   /** Stores context-local state variables. */
   private static final ContextLocal<WorkArea> work = 
   new ContextLocal<WorkArea>()
   {
      protected synchronized WorkArea initialValue()
      {
         return new WorkArea();
      }
   };

   /** Layout manager. */
   private LayoutManager<O> layoutManager = null;

   /** Current (in focus) widget. */
   private Widget<O> current = null;

   /** Container insets. */
   private Insets insets = new Insets(0, 0, 0, 0);

   /** Container size. */
   private Dimension size = new Dimension(1, 1);

   /** Widgets stored in the container. */
   private ArrayList<Widget<O>>  widgets = new ArrayList<Widget<O>>();
   
   /** 
    * The set of widgets that are never selected for focus - they are ignored by 
    * {@linkplain #checkWidget(Widget, boolean)}.
    */
   private Set<Widget<?>> ignoreFocusSet; 
   
   /**
    * Constructor.
    */
   public AbstractContainer()
   {
   }

   /**
    * Add widget to container.
    *
    * @param   widget
    *          Widget to add.
    *
    * @return  <code>this</code> for fluent syntax support.
    */
   @Override
   public Container<O> add(Widget<O> widget)
   {
      widgets.add(widget);

      widget.setParent(this);

      return this;
   }

   /**
    * This method should be overridden in a container that allows clipping.
    *
    * @return  <code>true</code> if container clips its children.
    */
   @Override
   public boolean clipping()
   {
      return false;
   }

   /**
    * Get widget which is in focus.
    *
    * @return  focused widget.
    */
   @Override
   @SuppressWarnings("unchecked")
   public Widget<O> currentFocus()
   {
      if (current == null)
      {
         return null;
      }
      
      // pop-up menu
      int popupMenuId = current.config() == null ? -1 : current.config().popupMenuId;
      if (popupMenuId != -1)
      {
         Menu<O> popupMenu = (Menu<O>) current.screen().getRegistry().getComponent(popupMenuId);
         if (popupMenu.config() != null && popupMenu.config().visible)
         {
            return popupMenu.currentFocus();
         }
      }
      
      if (!(current instanceof Container))
      {
         return current;
      }
      
      if (current instanceof SubMenu)
      {
         SubMenu<?> sm = (SubMenu<?>) current;
         if (!sm.config().displayed || sm.currentFocus() == null)
         {
            return current;
         }
      }

      return ((Container<O>) current).currentFocus();
   }

   /**
    * Get default focusable widget.
    *
    * @return  default focusable widget.
    */
   @Override
   public Widget<O> defaultFocus()
   {
      Widget<O> defFocus = null;

      for (Widget<O> c : widgets)
      {
         if (c.focusTraversable())
         {
            defFocus = c;
            break;
         }
      }

      if (defFocus == null)
         throw new IllegalStateException("no focus-traversable widgets found");

      if (defFocus instanceof Container)
         return ((Container<O>) defFocus).defaultFocus();

      return defFocus;
   }

   /**
    * Get widget dimension.
    *
    * @return  widget dimension.
    */
   @Override
   public Dimension dimension()
   {
      return new Dimension(width(), height());
   }

   /**
    * Lay out container.
    */
   @Override
   public void doLayout()
   {
      if (layoutManager != null)
         layoutManager.doLayout(this);
   }
   
   /**
    * Find the widget positioned just bellow the specified mouse position (in physical units).
    * Each container should search for own widgets. Because widgets are relative positioned 
    * inside containers the container physical position is adjusted before search.  
    * 
    * @param    p
    *           The mouse physical location.
    * 
    * @return   A widget over which the mouse is positioned or <code>null</code> if none found.
    */
   @Override
   public Widget<O> findMouseSource(NativePoint p)
   {
      // to allow z-order awareness, process the widgets in reverse order: the top-most widgets
      // are always last in the widget list
      for (int i = widgets.size() - 1; i >= 0; i--)
      {
         Widget<O> w = widgets.get(i);
         
         if (!w.isVisible() || !w.physicalBounds().contains(p))
         {
            continue;
         }
         
         Widget<O> found = w.findMouseSource(p.minus(w.physicalLocation()));
         return (found == null) ? w : found;
      }

      return null;
   }
   
   /**
    * Register container widgets.  Use to register frame widgets 
    * 
    * @param   reg
    *          Widgets registry
    */
   public void registerWidgets(WidgetRegistry<O> reg)
   {
      for (Widget<O> w : widgets)
      {
         if (w instanceof AbstractContainer)
         {
            ((AbstractContainer<O>) w).registerWidgets(reg);
         }
         if (w.config() != null && w.config().id != null)
         {
            reg.addWidget(w);
         }
      }
   }
   
   /** 
    * Draw widget.
    */
   @SuppressWarnings("unchecked")
   @Override
   public void draw()
   {
      if (!isVisible())
         return;

      O screen = screen();
      Widget<O> parent = parent();
      
      List<Widget<O>> drawableWidgets = widgets;
      if (parent instanceof TitledWindow)
      {
         drawableWidgets = getDrawableWidgets();
      }
      
      int windowId = window().config().id.getId();
      
      // old method for nested containers
      for (Widget<O> c : prepareWidgets(drawableWidgets))
      {
         if (c.isVisible())
         {
            Rectangle bounds = c.bounds().translate(childrenLocation(c));
            List<Rectangle> clippings = screen.getClippings(windowId, bounds);
            if (!clippings.isEmpty())
            {
               c.draw();
            }

            // mark the label widget as "was visible" because we are going to show it
            if (c instanceof LabeledWidget)
            {
               ((LabeledWidget<O>) c).setWasVisible();
            }
         }
      }
   }

   /**
    * Get children location for one child.
    * 
    * @param   child
    *          Child component.
    *          
    * @return  Point - parents location
    */
   public Point childrenLocation(Widget<O> child)
   {
      return child.parent().screenLocation();
   }

   /**
    * Indicates the ending of the resize process. All repainting between
    * <code>startSizeChange</code> and <code>endSizeChange</code> calls is
    * suppressed.
    *
    * @param    repaint
    *           If <code>true</code> then the current bounding rectangle and
    *           the rectangle saved during <code>startSizeChange</code> call
    *           are repainted.
    */
   @Override
   public void endSizeChange(boolean repaint)
   {
      WorkArea wa = work.get();
      
      Rectangle savedSize = wa.savedSizes.pop();

      if (repaint)
      {
         if (savedSize != null)
            repaint(savedSize);
         
         repaint();
      }
   }

   /**
    * Returns the focused direct child Component.
    *
    * @return   The focused direct child Component.
    */
   @Override
   @SuppressWarnings("unchecked")
   public Widget<O> focus()
   {
      if (config() != null && config().popupMenuId != -1)
      {
         Menu<O> menu = (Menu<O>) screen().getRegistry().getComponent(config().popupMenuId );
         if (menu != null && menu.isEnabled() && menu.isVisible())
         {
            return menu;
         }
      }
      
      return current;
   }

   /**
    * Check if widget is focus traversable.
    *
    * @return  <code>true</code> if widget is focus traversable.
    */
   @Override
   public boolean focusTraversable()
   {
      if (!super.focusTraversable())
         return false;

      return hasFocusable(null);
   }

   /**
    * Get widget for provided index or <code>null</code> if no such component
    * exists.
    *
    * @param   index
    *          Widget index.
    *
    * @return  widget at specified index or <code>null</code> if index is out
    *          of range.
    */
   @Override
   public Widget<O> get(int index)
   {
      if (index < 0 || index >= widgets.size())
         return null;

      return widgets.get(index);
   }

   /**
    * Get clipping rectangle.
    *
    * @return   clipping rectangle.
    */
   @Override
   public Rectangle getClipRect()
   {
      return new Rectangle(new Point(0, 0), dimension());
   }

   /**
    * Reports if there is at least 1 focusable widget that is not the given
    * widget.
    *
    * @param    exclude
    *           A widget to ignore or <code>null</code> if any focusable
    *           widget will do.
    *
    * @return   <code>true</code> if there is at least one other widget that
    *           can be focused.
    */
   @Override
   public boolean hasFocusable(Widget<O> exclude)
   {
      for (Widget<O> widget : widgets)
      {
         if (widget == exclude)
            continue;

         if (widget.focusTraversable())
            return true;
      }

      return false;
   }

   /**
    * Get widget height.
    *
    * @return  widget height.
    */
   @Override
   public double height()
   {
      return size.height;
   }

   /**
    * Get container insets.
    *
    * @return  container insets.
    */
   @Override
   public Insets insets()
   {
      return insets;
   }

   /**
    * Get widget minimum size.
    *
    * @return  widget minimum size.
    */
   @Override
   public Dimension minimumSize()
   {
      if (layoutManager == null)
         return size;

      return layoutManager.minimumSize(this);
   }

   /**
    * Move specified widget before the widget "before".
    * Widgets must be in the same field-group, that is checked on the server-side.
    * 
    * @param   widget
    *          Widget to move.
    * @param   before
    *          The widget before to insert mover.
    */
   @Override
   public void moveBeforeTabItem(Widget<O> widget, Widget<O> before)
   {
      if (widgets.size() == 0)
         return;

      if (!widgets.contains(widget))
      {
         throw new RuntimeException("widgets do not contain specified: " + before);
      }
      
      if (!widgets.contains(before)) 
      {
         throw new RuntimeException("widgets do not contain specified: " + before);
      }

      widgets.remove(widget);
      widgets.add(widgets.indexOf(before), widget);
   }

   /**
    * Mode specified widget to the bottom of the widget list.
    *
    * @param   widget
    *          Widget to mover.
    */
   @Override
   public void moveToBottom(Widget<O> widget)
   {
      moveToZpos(widget, 0);
   }

   /**
    * Move specified widget to the top of widget list.
    *
    * @param   widget
    *          Widget to mover.
    */
   @Override
   public void moveToTop(Widget<O> widget)
   {
      int size = widgets.size();

      if (size == 0)
         return;

      if (!widgets.contains(widget))
         return;

      if (widgets.get(size - 1).equals(widget))
         return;

      widgets.remove(widget);
      widgets.add(widget);
   }

   /**
    * Move widget to specified position in the widget list. For overlaying
    * widgets this is basically position in widget Z-order.
    *
    * @param   widget
    *          Widget to mover.
    * @param   pos
    *          Position in Z-order.
    */
   @Override
   public void moveToZpos(Widget<O> widget, int pos)
   {
      if (pos < widgets.size() && widgets.contains(widget))
      {
         int ndx = widgets.indexOf(widget);

         if (ndx == pos)
            return;

         if (ndx < pos)
            pos--;

         widgets.remove(widget);
         widgets.add(pos, widget);
      }
   }

   /**
    * Select next focusable widget.
    */
   @Override
   public void nextFocus()
   {
      focusWorker(true);
   }

   /**
    * Select previous focusable widget.
    */
   @Override
   public void prevFocus()
   {
      focusWorker(false);
   }

   /**
    * Process key event.
    *
    * @param   event
    *          Key event to process.
    */
   @Override
   public void processKeyEvent(KeyInput event)
   {
      super.processKeyEvent(event);

      if (event.isConsumed())
         return;
      
      // TODO process pop-up menu?

      if (current != null && current != this)
         current.processKeyEvent(event);
   }

   /**
    * Remove specified widget from the widget list.
    * <p>
    * The widget will also be destroyed.
    * 
    * @param   widget
    *          Widget to remove.
    */
   @Override
   public void remove(Widget<O> widget)
   {
      detach(widget);
      widget.destroy();
   }
   
   /**
    * Detach this widget from this container.  The widget will not be destroyed.
    * 
    * @param   widget
    *          Widget to detach.
    */
   @Override
   public void detach(Widget<O> widget)
   {
      widgets.remove(widget);
      // the widget's parent needs to be removed; downstream logic (some of it related to focus
      // processing) is dependent of the widget's parent being set or not.
      widget.setParent(null);

      if (current == widget)
      {
         current = null;
         current = currentFocus();
      }
   }
   
   /**
    * Destroy this widget.  Notify all of its components that the widget has been destroyed.
    */
   public void destroy()
   {
      for (Widget<O> w : widgets)
      {
         w.destroy();
      }

      super.destroy();
   }

   /**
    * Request widget repainting.
    */
   @Override
   public void repaint()
   {
      WorkArea wa = work.get();
      
      if (wa.savedSizes.isEmpty())
         super.repaint();
   }

   /**
    * Replace specified widget with another with preserving widget Z-order.
    *
    * @param   oldWidget
    *          Old widget.
    * @param   newWidget
    *          New widget.
    */
   @Override
   public void replace(Widget<O> oldWidget, Widget<O> newWidget)
   {
      if (oldWidget != null)
      {
         widgets.remove(oldWidget);
         oldWidget.destroy();
         oldWidget.setParent(null);

         if (current == oldWidget)
            current = newWidget;
      }

      add(newWidget);
   }

   /**
    * Request focus.
    */
   @Override
   public void requestFocus()
   {
      Widget<O> focus = currentFocus();

      if (focus == null)
      {
         focus = defaultFocus();
      }
      
      if (focus == this)
      {
         super.requestFocus();
      }
      else if (focus != null)
      {
         if (this instanceof SubMenu)
         {
            super.requestFocus();
         }
         focus.requestFocus();
      }
   }

   /**
    * Set focus to specified widget.
    *
    * @param   focus
    *          Widget to set focus on.
    */
   @Override
   public void setFocus(Widget<O> focus)
   {
      setFocusInt(focus);
      if (parent() != null)
         parent().setFocus(this);
   }

   /**
    * Set container height.
    *
    * @param   height
    *          New container height.
    */
   @Override
   public void setHeight(double height)
   {
      size.height = Coordinate.scale(height);
   }

   /**
    * Set container insets.
    *
    * @param   insets
    *          Insets to set.
    */
   @Override
   public void setInsets(Insets insets)
   {
      this.insets = Coordinate.scale(insets);
   }

   /**
    * Set layout manager for the container.
    *
    * @param   manager
    *          New layout manager.
    */
   @Override
   public void setLayout(LayoutManager<O> manager)
   {
      this.layoutManager = manager;
   }

   /**
    * Get active layout manager for the container.
    *
    * @return  Active layout manager.
    */
   @Override
   public LayoutManager<O> getLayout()
   {
      return layoutManager;
   }

   /**
    * Set container size.
    *
    * @param   size
    *          Size to set.
    */
   @Override
   public void setSize(Dimension size)
   {
      setSize(size.width, size.height);
   }

   /**
    * Set container size.
    *
    * @param   width
    *          New width.
    * @param   height
    *          New height.
    */
   @Override
   public void setSize(double width, double height)
   {
      if (isVisible())
      {
         repaint();
      }

      size = new Dimension(width, height);

      if (isVisible())
      {
         repaint(); // post repainting of the new rectangle
      }
   }
   
   /**
    * Set container width.
    *
    * @param   width
    *          Width to set.
    */
   @Override
   public void setWidth(double width)
   {
      size.width = Coordinate.scale(width);
   }

   /**
    * Indicates the beginning of the resize process. All repainting between
    * <code>startSizeChange</code> and <code>endSizeChange</code> calls is
    * suppressed.
    *
    * @param    saveOldRect
    *           If <code>true</code> then the current bounding rectangle is
    *           saved for future repainting.
    */
   @Override
   public void startSizeChange(boolean saveOldRect)
   {
      WorkArea wa = work.get();
      
      if (saveOldRect)
      {
         wa.savedSizes.push(bounds());
      }
      else
      {
         wa.savedSizes.push(null);
      }
   }

   /**
    * Get all widgets stored in the container.
    *
    * @return  array of stored widgets.
    */
   @SuppressWarnings("unchecked")
   @Override
   public Widget<O>[] widgets()
   {
      return widgets.toArray(new Widget[widgets.size()]);
   }

   /**
    * Get all widgets stored in the container.
    *
    * @return  the new list with stored widgets.
    */
   public List<Widget<O>> getWidgetsList()
   {
      return Collections.unmodifiableList(widgets);
   }
   
   /**
    * Get widget width.
    *
    * @return  widget width.
    */
   @Override
   public double width()
   {
      return size.width;
   }

   /**
    * Set focus to first focusable widget.
    */
   @Override
   public void firstFocus()
   {
      Widget<O> widget = locateFocusTraversable(widgets);

      if (widget instanceof Container)
         ((Container<O>) widget).firstFocus();

      current = widget;
   }

   /**
    * Set focus to last focusable widget.
    */
   @Override
   public void lastFocus()
   {
      Widget<O> widget = locateFocusTraversable(Iterables.reverse(widgets));

      if (widget instanceof Container)
         ((Container<O>) widget).lastFocus();

      current = widget;
   }
         
   /**
    * Calculate container virtual size.  Virtual size is calculated depending on location and
    * size of container widgets. 
    * 
    * @return  see above.
    */
   public Dimension virtualSize()
   {
      double width = 0, height = 0;

      for (Widget<O> widget : widgets)
      {
         if (widget.isVisible())
         {
            Point origin = widget.location();
            Dimension dimension = (widget instanceof Frame) ? 
                     ((Frame<?>) widget).getSavedDimension() : widget.dimension();
            // virtual width
            double vw = origin.x + dimension.width;
            if (vw > width)
            {
               width = vw;
            }
            // virtual height
            double vh = origin.y + dimension.height;
            if (vh > height)
            {
               height = vh;
            }
         }
      }

      return new Dimension(width, height);
   }
   
   /** 
    * Returns the set of widgets that are never selected for focus.
    * 
    * @return  set of widgets or <code>null</code>. 
    */
   @Override
   public Set<Widget<?>> getIgnoreFocus()
   {
      return Collections.unmodifiableSet(ignoreFocusSet);
   }
   
   /**
    * Sets the set of widgets that are never selected for focus.
    * <code>null</code> may be used to clear the set.
    * 
    * @param   widgets
    *          Set of widgets or <code>null</code>.
    */
   @Override
   public void setIgnoreFocus(Collection<Widget<?>> widgets)
   {
      if (widgets == null || widgets.isEmpty())
      {
         ignoreFocusSet = null;
         return;
      }
      
      ignoreFocusSet = new HashSet<>();
      ignoreFocusSet.addAll(widgets);
   }
   
   /**
    * Set focus to specified widget.
    *
    * @param   focus
    *          Widget to set focus on.
    */
   protected void setFocusInt(Widget<O> focus)
   {
      current = focus;
   }
   
   /**
    * Prepare widgets before drawing
    * 
    * @param   widgets
    *          Container children.
    *          
    * @return  widgets, prepared for drawing.
    */
   protected List<Widget<O>> prepareWidgets(List<Widget<O>> widgets)
   {
      return Collections.unmodifiableList(widgets);
   }
   
   /**
    * Add widget directly to the list without triggering repaint or other
    * operations.
    *
    * @param   widget
    */
   protected void addDirectly(Widget<O> widget)
   {
      widgets.add(widget);
   }

   /**
    * Check if specified widget belongs to this container.
    *
    * @param   widget
    *          Widget to check.
    *
    * @return  <code>true</code> if widget belongs to container.
    */
   protected boolean belongsTo(Widget<O> widget)
   {
      return widgets.contains(widget);
   }

   /**
    * Build clipped rectangle for given coordinates.
    * 
    * @param   origin
    *          Physical location.
    * @param   dimension
    *          Physical dimension.
    * @param   children
    *          Flag to intersect the rectangle with the to-be-drawn chidren.
    *          
    * @return  see above.
    */
   protected NativeRectangle clipRectangle(NativePoint     origin,
                                           NativeDimension dimension, 
                                           boolean         children)
   {
      NativeRectangle nclip = super.clipRectangle(origin, dimension);
      
      if (children)
      {
         O screen = screen();
         Widget<O> parent = parent();
         
         List<Widget<O>> drawableWidgets = widgets;
         if (parent instanceof TitledWindow)
         {
            drawableWidgets = getDrawableWidgets();
         }
         
         int windowId = window().config().id.getId();
         CoordinatesConversion coords = screen.coordinates();
         
         Rectangle drawing = null;
         
         // collect the clipping rectangles from all affected children and merge them
         for (Widget<O> c : drawableWidgets)
         {
            if (c.isVisible())
            {
               Point screenLoc = childrenLocation(c);
               Rectangle bounds = c.bounds().translate(screenLoc);
               List<Rectangle> clippings = screen.getClippings(windowId, bounds);
               if (!clippings.isEmpty())
               {
                  // inform the driver to draw only into the specified rectangle
                  Rectangle boundary = null;
                  for (Rectangle clip : clippings)
                  {
                     boundary = (boundary == null ? clip : boundary.union(clip));
                  }

                  boundary = boundary.translate(-screenLoc.x, -screenLoc.y);
                  
                  drawing = (drawing == null ? boundary : drawing.union(boundary));
               }
            }
         }

         nclip = nclip.intersection(coords.rectangleToNative(drawing));
      }

      return nclip;
   }

   /**
    * Helper method for traversing widgets in the container.
    *
    * @param   iter
    *          Iterable for the contained widgets.
    *
    * @return  Found widget or <code>null</code> if no matches found.
    */
   private Widget<O> locateFocusTraversable(Iterable<Widget<O>> iter)
   {
      for (Widget<O> widget : iter)
      {
         if (widget.focusTraversable())
            return widget;
      }

      return null;
   }

   /**
    * Search for the next or previous focusable widget in the container and
    * set focus on it.
    *
    * @param   direction
    *          Widget search direction.
    */
   protected void focusWorker(boolean direction)
   {
      Iterable<Widget<O>> iter = (direction) ?
                                 Iterables.directFrom(widgets, current) :
                                 Iterables.reverseFrom(widgets, current);

      if (iter == null)
         throw new IllegalStateException("Focused widget does not belong to container");

      for (Widget<O> widget : iter)
      {
         if (checkWidget(widget, direction))
            return;
      }

      if (parent() != null)
      {
         if (direction)
             parent().nextFocus();
         else
             parent().prevFocus();
         return;
      }

      iter = (direction) ? widgets : Iterables.reverse(widgets);

      for (Widget<O> widget : iter)
      {
         if (widget == current)
         {
            current = null;
            return;
         }

         if (checkWidget(widget, direction))
            return;
      }
   }

   /**
    * Check potential focus candidate and set focus on it if it matches
    * criteria.
    *
    * @param   widget
    *          Candidate widget.
    * @param   direction
    *          Search direction.
    *
    * @return  <code>true</code> if widget was focused and <code>false</code>
    *          otherwise.
    */
   protected boolean checkWidget(Widget<O> widget, boolean direction)
   {
      if (!widget.focusTraversable())
         return false;

      // check whether an explicit ignore was set for the widget
      if (ignoreFocusSet != null && ignoreFocusSet.contains(widget))
      {
         return false;
      }
      
      if (widget instanceof Container)
      {
         if (direction)
            ((Container<O>) widget).firstFocus();
         else
            ((Container<O>) widget).lastFocus();
      }

      widget.requestFocus();

      return true;
   }

   /**
    * Get the widgets which can be drawn, from this container.
    * 
    * @return   All the widgets which touch the clipping areas, according to their z-order.
    */
   private List<Widget<O>> getDrawableWidgets()
   {
      Widget<O>[] widgets = widgets();
      int lim = widgets.length;
      int from = 0;
      Widget<O> c = null;
   
      O screen = screen();
      
      int windowId = window().config().id.getId();

      // new method for the top container
      // search down the widgets until all clipped widgets are
      // touched; from those, return the last found frame.
      screen.startScreenSearch(windowId);
      int lastFrom = lim - 1;
      boolean foundContainer = false;
   
      for (from = lim - 1; from >= 0; from--)
      {
         c = widgets[from];
   
         if (!c.isVisible())
         {
            continue;
         }
   
         if (!(c instanceof Container))
         {
            if (screen.markArea(windowId, c.bounds()) && screen.getInvalidChars(windowId) == 0)
            {
               from = lastFrom;
               break;
            }
            continue;
         }
   
         // in redirected mode, only the topmost frame is of interest
         if (screen.isRedirected())
            break;
   
         if (screen.markArea(windowId, c.bounds()))
         {
            lastFrom = from;
            foundContainer = true;
            if (screen.getInvalidChars(windowId) == 0)
            {
               break;
            }
         }
      }
   
      if (from < 0)
      {
         if (foundContainer && screen.getInvalidChars(windowId) > 0)
         {
            // if there are clipped characters left and we found a
            // container, then drawing must start from the last found
            // position
            from = lastFrom;
         }
         else
         {
            from = 0;
         }
      }
   
      List<Widget<O>> res = new ArrayList<>();
      for (int i = from; i < lim; i++)
      {
         res.add(widgets[i]);
      }
      
      return res;
   }

   /** 
    * Stores global data relating to the state of the current context.
    */
   private static class WorkArea
   {  
      /** Container saved sizes stack. */       
      private Stack<Rectangle> savedSizes = new Stack<Rectangle>();
   }
}
